"""Provides an Engine for measuring a dataset.

This module defines the `MeasurementEngine` class, which facilitates the 
measurement of data using various measurement types and factories. It also 
includes helper functions to validate measurement types.

TODO: Get the scenario logger and set it as the logger for the engine!
"""

from collections.abc import Iterable
import inspect
from logging import getLogger
from typing import Any, Literal, Optional, Union, cast, get_args, get_origin
import uuid
import numpy as np
from adgtk.common.defaults import SCENARIO_LOGGER_NAME
from adgtk.common import UnableToMeasureException
from adgtk.tracking import ExperimentRunFolders, MetricTracker
from .factory import create_measurement
from .factory import (
    ClassBasedComparison,
    ClassBasedMeasurement,    
    MeasFactoryEntry,
    direct_comparison,    
    direct_measurement,
    distribution_measurement,
    distribution_comparison,
    get_measurements_by_tag,
    get_measurements_by_type,
    get_measurement_factory_entry,
    measurement_type,
    supports_factory
)

# ----------------------------------------------------------------------
# Development support only
# ----------------------------------------------------------------------
DEBUG = False

# ----------------------------------------------------------------------
# Structure
# ----------------------------------------------------------------------
calculation_type = Literal["avg", "sum", "max", "min", "raw", "distribution"]


# ----------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------
def supports_measurement_type(func, *args) -> bool:
    """
    Validates if the provided arguments match the expected types of a function.

    Args:
        func: The function whose parameter types are to be validated.
        *args: The arguments to validate against the function's parameter types.

    Returns:
        bool: True if all arguments match the expected types, False otherwise.
    """
    sig = inspect.signature(func)
    parameters = sig.parameters
    for i, (_, param) in enumerate(parameters.items()):
        if param.annotation != inspect.Parameter.empty:
            expected_type = param.annotation
            if i < len(args):
                arg = args[i]
                if DEBUG:
                    # Show the expected type and the argument type
                    print(f"Expected type: {expected_type}, "
                          f"Argument type: {type(arg)}")
                # Handle Union types
                if get_origin(expected_type) is Union:
                    if not any(isinstance(arg, t) for t in get_args(expected_type)):
                        return False
                # Handle regular types
                elif not isinstance(arg, expected_type):
                    return False

    if len(args) != len(parameters):
        return False
    
    return True
# ----------------------------------------------------------------------
# Engine
# ----------------------------------------------------------------------


class MeasurementEngine:
    """Drives measurements of data.

    The `MeasurementEngine` class manages the registration of measurement 
    factories, tracks metrics, and performs measurements on datasets.

    Attributes:
        engine_id (Optional[str]): A unique identifier for the engine.
        measurements (dict[str, supports_factory]): A dictionary of registered 
            measurement factories.
        details (dict[str, MeasFactoryEntry]): A dictionary containing details 
            about each registered measurement factory.
        metric_tracker (MetricTracker): Tracks metrics for the measurements.
        logger: Logger instance for logging engine-related events.
    """

    def __init__(
        self,
        engine_id: Optional[str] = None,
        add_factory_ids: Optional[list[str]] = None,
        add_by_type: Optional[measurement_type] = None,
        add_by_tag: Optional[Union[str, list[measurement_type]]] = None
    ) -> None:
        """
        Initializes the MeasurementEngine.

        Args:
            engine_id (Optional[str]): A unique identifier for the engine.
            add_factory_ids (Optional[list[str]]): A list of factory IDs to 
                register during initialization.
            add_by_type (Optional[measurement_type]): A measurement type to 
                register factories by.
            add_by_tag (Optional[Union[str, list[measurement_type]]]): A tag 
                or list of tags to register factories by.
        """
        self.engine_id = engine_id or str(uuid.uuid4())
        self.measurements: dict[str, supports_factory] = {}
        self.details: dict[str, MeasFactoryEntry] = {}
        self.metric_tracker = MetricTracker(
            name=self.engine_id,
            purpose="measurement"
        )
        self.logger = getLogger(SCENARIO_LOGGER_NAME)

        if add_factory_ids is not None:
            for entry in add_factory_ids:
                self.add(entry)
        if add_by_type is not None:
            entries = get_measurements_by_type(add_by_type)
            for entry in entries:
                self.add(entry['factory_id'])
        if add_by_tag is not None:
            entries = get_measurements_by_tag(add_by_tag)
            for entry in entries:                
                self.add(entry['factory_id'])

    def clear_results(self) -> None:
        """Clears all prior measurements but retains the definitions.
        """
        self.metric_tracker.clear_results()

    def get_all_data(self, label:str) -> list:
        """Gets all the data stored for a label

        Args:
            label (str): The label of the measurement

        Returns:
            list: the data
        """
        return self.metric_tracker.get_all_data(label)
    
    def measurement_count(self, label:str)-> int:
        """The number of measurements stored

        Args:
            label (str): the label of the measurement

        Returns:
            int: The count of data stored
        """
        return self.metric_tracker.measurement_count(label)

    def get_average(self, label:str) -> float:
        """Returns the average value for a measurement

        Args:
            label (str): the label of the measurement

        Returns:
            float: The average value
        """
        return self.metric_tracker.get_average(label)

    def get_latest_value(self, label:str)-> float:
        """Gets the latest value recorded

        Args:
            label (str): the label of the measurement

        Returns:
            float: The latest value
        """
        return self.metric_tracker.get_latest_value(label)

    def get_latest_distribution(self, label:str) -> np.ndarray:
        """Gets the latest distribution stored with a label.

        Args:
            label (str): the label of the measurement

        Returns:
            np.ndarray: The latest distribution
        """
        return self.metric_tracker.get_latest_distribution(label)

    def get_description(self, factory_id:str) -> str:
        """Retrieves the description of a measurement.

        Args:
            factory_id (str): The measurement to retrieve

        Returns:
            str: The description
        """
        if factory_id not in self.details.keys():
            raise IndexError("Unable to locate id %s", factory_id)
        
        entry = self.details[factory_id]
        return entry["description"]

        
    def add(self, factory_id: str, **kwargs) -> None:
        """
        Registers a measurement factory with the engine.

        Args:
            factory_id (str): The ID of the factory to register.
            **kwargs: Additional arguments for the factory creation.

        Raises:
            IndexError: If the factory ID is invalid or not found.
        """
        try:
            entry = get_measurement_factory_entry(factory_id)
            self.measurements[factory_id] = create_measurement(factory_id)
            self.details[factory_id] = entry
            self.metric_tracker.register_metric(label=factory_id)
        except IndexError:
            self.logger.error(
                "Measurement engine %s unable to add %s due to already "
                "registered. Ignoring request",
                self.engine_id, factory_id)

    def _update_tracker(
        self,
        label: str,
        results: list,
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Records the results of a measurement in the tracker.

        Args:
            label (str): The tracker label.
            results (list): The data to add.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """
        if len(results) == 0:
            results = [0]
            self.logger.warning(
                "Measurement engine %s performed zero measurements.",
                self.engine_id)
        if record_as == "avg":
            value = sum(results) / len(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "max":
            value = max(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "min":
            value = min(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "sum":
            value = sum(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "raw":
            self.metric_tracker.add_raw_data(label=label, values=results)      

    def measure(
        self,
        data: Iterable,
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Performs measurements on the provided dataset.

        Args:
            data (Iterable): The dataset to measure.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """        
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedMeasurement, meas)
            else:
                meas = cast(direct_measurement, meas)     

            # first, does the measurement want all the data?
            if supports_measurement_type(meas, data):
                try:
                    result = meas(data)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass
            else:           
                # if not, then iterate over the values
                # this is a fallback. measurements should be designed to
                # consider iterable values.
                for entry in data:
                    # Verify if the measurement type is supported
                    try:
                        if supports_measurement_type(meas, entry):                    
                            result = meas(entry)
                            if isinstance(result, (int, float)):
                                all_results.append(result)
                            elif isinstance(result, list):
                                all_results.extend(result)
                        else:
                            self.logger.warning(
                                f"{self.engine_id} No valid data for {label}. "
                                "skipping measure")
                            break
                    except UnableToMeasureException:
                        # NO-OP
                        pass
            self._update_tracker(
                label=label, results=all_results, record_as=record_as)

    def measure_dataset_distribution(self, dataset: Iterable) -> None:
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedMeasurement, meas)
            else:
                meas = cast(distribution_measurement, meas)
            # now measure
            if supports_measurement_type(meas, dataset):
                try:
                    result = meas(dataset)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass            
            self._update_tracker(
                label=label,
                results=all_results,
                record_as="distribution"
            )

    def compare_dataset_distribution(
        self,
        dataset_one: Iterable,
        dataset_two: Iterable,
        record_as: calculation_type = "avg"
    ) -> None:
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedComparison, meas)
            else:
                meas = cast(distribution_comparison, meas)
            if supports_measurement_type(meas, dataset_one, dataset_two):
                try:
                    result = meas(dataset_one, dataset_two)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass            
            self._update_tracker(
                label=label,
                results=all_results,
                record_as=record_as
            )
            
    def compare(
        self,
        data: Iterable[tuple[Any, Any]],
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Performs measurements on the provided dataset.

        Args:
            data (Iterable): The dataset to measure.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedComparison, meas)
            else:
                meas = cast(direct_comparison, meas)                
            for a,b in data:
                # Verify if the measurement type is supported
                if supports_measurement_type(meas, a, b):                    
                    result = meas(a, b)
                    if isinstance(result, (int, float)):
                        all_results.append(result)
                    elif isinstance(result, list):
                        all_results.extend(result)
            self._update_tracker(
                label=label, results=all_results, record_as=record_as)
            
    def save_data(self, folders: ExperimentRunFolders) -> None:
        """Saves the data to disk using the pre-defined folder structure

        Args:
            folders (ExperimentRunFolders): The results folders
        """
        self.metric_tracker.save_data(folders)

    def report(self) -> None:
        """Prints a report to screen. Primary purpose is development.        
        """
        print(f"------- Measurement Engine: {self.engine_id} -------")
        max_key_length = max(len(k) for k in self.measurements.keys())
        for k, v in self.measurements.items():
            print(f"{k:<{max_key_length}}  {type(v).__name__}")

    def export_last_val_to_dict(self) -> dict:
        """Obtains the latest measurement as a dictionary.

        Returns:
            dict: the last value recorded by metric label
        """
        return self.metric_tracker.export_last_val_to_dict()



"""Pipeline provides a consistent set of tools for processing of data."""

import logging
from typing import Union, Protocol, runtime_checkable, Literal, Any
from dataclasses import dataclass
from adgtk.components import PresentableRecord, PresentableGroup


@dataclass
class PipelineConfig:
    """Configuration for the pipeline."""
    version: str
    debug: bool
    log_failures: bool


@dataclass
class PipelineRequest:
    """A request for the pipeline to work"""
    version: str
    data: Any


@runtime_checkable
class Pipeline(Protocol):
    """Provides a consistent set of tools for processing of data."""

    def __init__(self, config: PipelineConfig):
        """Initializes a pipeline.

        :param config: The pipeline configuration
        :type config: PipelineConfig
        """

    def process(self, request: PipelineRequest) -> Any:
        """Processes a pipeline request

        :param request: The request to process
        :type request: PipelineRequest
        :return: The processed data
        :rtype: Any
        """



"""Contains the text for when no flag is set"""


intro = f"""
======================================
ADGTK - Agent Data Generation Tool-Kit
======================================

Interaction with ADGTK is done through the command line. It should be run from the root directory of an experiment folder or used to create a new project.

You can use the wizard to create an experiment or hand craft your experiment with YAML or TOML.


For more information please visit the project at http://todo/docs <= UPDATE LINK!

Key definitions
===============

    - project    : A set of folders that contains the experiments and results.
    - experiment : A single definition and associated results from that definition.
    - blueprint  : A set of configurations used to build an experiment and is registered in the factory.
    - sample     : Provides a sample project, a "helloworld" project.


Usage
=====

adgtk-mgr [options]  

Commands:
   project      : Project management (create, destroy)
   experiment   : Experiment operations (create, run, list, preview)
   factory      : List Factory or if include a group the group listing
     
options:
  -h, --help            show a help message and exits
  -f FILE, --file FILE  override the settings file with this file
  --version             show program's version number and exit
  --yaml                Use YAML format when creating the project settings file


Project
-------
   $ adgtk-mgr project create example1   Creates a new project called example1
   $ adgtk-mgr project destroy example1  deletes the example1 project

Experiment
----------
   $ adgtk-mgr experiment list           lists all available experiments
   $ adgtk-mgr experiment create         Starts a wizard to build an experiment
   $ adgtk-mgr experiment create exp1     Starts a wizard to build an experiment with the name exp1
   $ adgtk-mgr experiment run            via a menu select and run an experiment
   $ adgtk-mgr experiment run exp1       Run exp1

Factory
-------
   $ adgtk-mgr factory                   Lists available factory blueprints
   $ adgtk-mgr factory agent             Lists agent factory blueprints
   
Modify the start example
------------------------
(Loads settings1.toml and runs the factory command)

   $ adgtk-mgr --file settings1.toml factory   
"""



# pylint: disable=pointless-string-statement
"""The primary purpose of this module is to support the built-in
components. Secondarly the framework itself can use this for processing.
The overall concept is not to fully implement a database but rather
provide a Protocol for any implementation of a wrapper for a database.

By using a consistent wrapper it enables ease of extensibility as well
as code maintenance as it allows for loose coupling across Modules.
"""

# pylint: disable=unused-argument
from __future__ import annotations
from enum import Enum, auto
from typing import (
    Union,
    Protocol,
    runtime_checkable,
    Any,
    Iterable)
from adgtk.components.data.records import PresentableRecord

# ----------------------------------------------------------------------
# Types and Protocols
# ----------------------------------------------------------------------


class SearchOption(Enum):
    """Search capabilities for a data store"""
    NONE = auto()


@runtime_checkable
class RecordStore(Protocol):
    """Provides a consistent way for the built-in objects to interact
    with a data store such as vector database. As most database systems
    differ in their implementations this protocol provides a wrapper
    pattern for any potential use when built-in components are used."""

    # -----------------------------------------------------------------
    # Required methods
    # -----------------------------------------------------------------

    def insert(self, record: PresentableRecord) -> None:
        """Insert a single record

        :param record: _description_
        :type record: PresentableRecord
        """

    def bulk_insert(self, records: Iterable[PresentableRecord]) -> None:
        """insert several records at once

        :param records: _description_
        :type records: Iterable[PresentableRecord]
        """

    def clear_all_records(self) -> None:
        """Clears all records from the system        
        """

    def __len__(self) -> int:
        """The number of the records in the store

        :return: The count of records
        :rtype: int
        """


@runtime_checkable
class CanFindRandomRecord(Protocol):
    """Can find random record"""

    def find_random_record(
        self,
        search_filters: Union[dict, None] = None,
        not_record: Union[PresentableRecord, None] = None,
        **kwargs
    ) -> list:
        """Finds a single record. If a search cannot use a filter
        then at most it should log the issue but not stop as long
        as it can return a record.

        Feature flag: can_get_random_record

        :param near_record: The record to compare against,
            defaults to None
        :type not_record: PresentableRecord
        :param search_filters: the search guidance
        :type search_filters: Union[dict, None]
        :return: a single random record if sufficient records exist
        :rtype: list
        """


@runtime_checkable
class CanRebuildFromDisk(Protocol):
    """Can rebuild from disk"""

    def rebuild_from_disk(self, filename: str) -> bool:
        """Replaces any current records with ones loaded from disk.

        :param filename: The filename with path
        :type filename: str
        :return: Success loading.
        :rtype: bool
        """


@runtime_checkable
class CanSaveToDisk(Protocol):
    """Can use the save_to_disk method w/filename"""

    def save_to_disk(self, filename: str) -> None:
        """Saves to disk all records. For wrapper classes this method
        should export from the database into a common format if the db
        itself does not implement this capability.

        :param filename: The filename with path
        :type filename: str
        """


@runtime_checkable
class CanShuffleRecords(Protocol):
    """Can use the shuffle method to shuffle records"""

    def shuffle(self) -> None:
        """Shuffles the order of records.
        Feature Flag: can_shuffle
        """


@runtime_checkable
class CanExportRecordsToDict(Protocol):
    """Can export records to Dict"""

    def export_to_dict(self, filters: Union[dict, None]) -> dict:
        """Exports the data based on the filters

        Feature Flag: can_export_to_dict

        :param filters: what if any should be used to filter the records
        :type filters: Union[dict, None]
        :return: a dict with requested data
        :rtype: dict
        """


@runtime_checkable
class CanImportRecordsToDict(Protocol):
    """Can import records to Dict"""

    def import_from_dict(
        self,
        data: dict,
        metadata: Union[dict[str, Any], None] = None
    ) -> bool:
        """Imports records from disk.

        Feature Flag: can_import_from_dict

        :param data: _description_
        :type data: dict
        :param metadata: _description_, defaults to None
        :type metadata: dict, optional
        :return: _description_
        :rtype: bool
        """


@runtime_checkable
class CanGetAllRecords(Protocol):
    """Can export records to Dict"""

    def get_all_records(self, as_copy: bool = True) -> list:
        """Exports all records in the store.

        feature flag: can_export_all_records_to_list

        :param as_copy: export as a new object, defaults to True
        :type as_copy: bool, optional
        :return: a list of all records
        :rtype: list
        """


@runtime_checkable
class CanSearchForSimilar(Protocol):
    """Can search for similar"""

    def search_for_similar(
        self,
        near_record: PresentableRecord,
        search_filters: Union[dict, None],
        **kwargs
    ) -> list:
        """Searches for a similar record. If a search cannot use a
        filter then at most it should log the issue but not stop as long
        as it can return one or more records.

        feature Flag: can_search_by_proximity

        :param near_record: The record to compare against
        :type near_record: PresentableRecord
        :param search_filters: the search guidance
        :type search_filters: Union[dict, None]
        :return: a list of one or more records.
        :rtype: list
        """


@runtime_checkable
class CanFindByTerm(Protocol):
    """Can find by term"""

    def find_by_term(self, term: str) -> list:
        """Searches not by record but by term.

        Feature Flag: can_search_by_term

        :param term: the term to search for
        :type term: str
        :return: a list of Records
        :rtype: list
        """



"""Handlers for the CLI tools for this project

# TODO: move logging into the individual files, ex. runner, builder, etc
  each module should have its own. ex. experiment running should use the
  unique name of the experiment, while the builder is a common one.

# TODO: set log-level via a flag. shared module w/running config

"""
from logging import Logger
import signal
from typing import Optional
import argparse
import sys
import os
from pydantic import ValidationError
import adgtk.factory.component as factory
from adgtk.tracking.structure import EXP_RESULTS_FOLDER
from adgtk.tracking.utils import build_project_folders
from adgtk.manager.constants import BOOT_PY, BOOT_FILENAME
from adgtk import __version__ as adgtk_ver
from adgtk.manager.intro import intro
import adgtk.experiment.builder as experiment_builder
import adgtk.experiment.runner as experiment_runner
import adgtk.tracking.project as project_manager
from adgtk.tracking.structure import AvailableExperimentModel
from adgtk.utils import create_logger, get_scenario_logger
# and for the local project bootstrap.py
import importlib.util

# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
PID_FILE = "adgtk-mgr.pid"


# ----------------------------------------------------------------------
# extensibility (getting user code into the ecosystem) - bootstrap.py
# ----------------------------------------------------------------------


def load_bootstrap():
    """Loads the user-defined bootstrap module if present."""
    bootstrap_path = os.path.join(os.getcwd(), "bootstrap.py")
    if not os.path.exists(bootstrap_path):
        print("No bootstrap.py found.")
        return None

    spec = importlib.util.spec_from_file_location(
        "user_bootstrap", bootstrap_path)
    if spec is None or spec.loader is None:
        raise ImportError("Unable to load bootstrap module")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

# ----------------------------------------------------------------------
# Bootstrap hook executor
# ----------------------------------------------------------------------


def execute_bootstrap_hooks(module):
    """Calls foundation(), builtin(), and user_code() from bootstrap.py."""
    for hook in ["foundation", "builtin", "user_code"]:
        func = getattr(module, hook, None)
        if callable(func):
            try:
                func()
            except ModuleNotFoundError as e:
                print(f"Missing module in bootstrap hook '{hook}': {e.name}")
                sys.exit(1)
            except Exception as e:
                print(f"Error in bootstrap hook '{hook}': {e}")
                sys.exit(1)


# ----------------------------------------------------------------------
# Signal management
# ----------------------------------------------------------------------
def cleanup(signum, frame):
    try:
        scenario_logger = get_scenario_logger()
        scenario_logger.info("signal recieved. exiting early")
    except RuntimeError:
        print("Signal recieved. exiting process")
        pass
    if os.path.exists(PID_FILE):
        os.remove(PID_FILE)
    sys.exit()
# ----------------------------------------------------------------------
# Module configuration
# ----------------------------------------------------------------------

# ----------------------------------------------------------------------
# Logging
# ----------------------------------------------------------------------
BUILDER_LOG = "adgtk.builder.log"
_logger: Optional[Logger] = None
# ----------------------------------------------------------------------
# Module variables
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Private
# ----------------------------------------------------------------------


def _parse_args() -> argparse.Namespace:
    """Parses the command line arguments

    Returns:
        argparse.Namespace: The command line input
    """
    parser = argparse.ArgumentParser()

    # modifiers
    parser.add_argument(
        '--version', action='version', version=f"ADGTK {adgtk_ver}")

    # now positional
    subparsers = parser.add_subparsers(
        dest="command", help="Available commands")

    # Project
    project_parser = subparsers.add_parser("project", help="Project actions")
    project_parser.add_argument(
        "action",
        choices=["create", "list"],
        help="Available project actions")
    project_parser.add_argument(
        "folder",
        type=str,
        nargs="?",
        help="The project folder name")

    # Experiment
    exp_parser = subparsers.add_parser(
        "experiment", help="Experiment actions")
    exp_parser.add_argument(
        "action",
        choices=["build", "list", "run"],
        help="Available experiment actions")
    exp_parser.add_argument(
        "--name",
        help="The name of the experiment",
        type=str)

    # Factory
    factory_parser = subparsers.add_parser(
        "factory", help="Factory actions")
    factory_parser.add_argument(
        "group",
        type=str,
        nargs="?",
        help="The group name")

    factory_parser.add_argument(
        "--tags",
        type=str,
        nargs='+',
        help="The tags to filter one"
    )

    # Batch
    batch_parser = subparsers.add_parser(
        "batch", help="Batch actions")
    batch_parser.add_argument(
        "action",
        choices=["run", "preview", "list", "create"],
        help="Available batch actions")
    batch_parser.add_argument(
        "name",
        type=str,
        nargs="?",
        help="The batch name")

    args = parser.parse_args()
    return args


# ---------------------------------------------------------------------
# Project creation
# ---------------------------------------------------------------------
def _create_boot_py(base_path: str = ""):
    """Creates the bootstrap.py file when creating a new project.

    :param base_path: The path to include, defaults to ""
    :type base_path: str, optional
    """
    file_w_path = os.path.join(base_path, BOOT_FILENAME)
    with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
        outfile.write(BOOT_PY)


def _create_project(name: str) -> None:
    """Creates a project

    :param name: The project/directory name to create
    :type name: str
    """
    print(f"Creating project: {name}")

    if os.path.exists(name):
        msg = f"Unable to create a project {name}. remove the other first!"
        print(msg)
    elif name is None:
        print("Please specify a name of your project")
    else:
        print(f"Attempting to create {name}")

        os.makedirs(name, exist_ok=True)
        os.chdir(name)
        build_project_folders(base_path=os.getcwd())
        _create_boot_py(base_path=os.getcwd())
        # and let the user know it worked
        print(f"Successfully created project {name}")
        print(f"`cd {name}` to work with the project")
        return None

    print(f"Error creating {name}")


def _list_projects() -> None:
    """Experience in which a list of available projects are sent to the
    screen.
    """
    project_folders = []
    folders = os.listdir("./")
    for folder_or_file in folders:
        if os.path.isdir(folder_or_file):
            if is_project(folder_or_file):
                project_folders.append(folder_or_file)

    print("Available projects")
    print("==================")
    for project in project_folders:
        print(f"- {project}")


def _run_bootstrap():
    """Loads the bootstrap file
    """
    # invoke the expected functions
    try:
        bootstrap = load_bootstrap()
        if bootstrap and hasattr(bootstrap, 'foundation'):
            bootstrap.foundation()
        if bootstrap and hasattr(bootstrap, 'builtin'):
            bootstrap.builtin()
        if bootstrap and hasattr(bootstrap, 'user_code'):
            bootstrap.user_code()

    except Exception as e:
        print(f"Unable to load boostrap. Code raised: {e}")
        sys.exit(1)


def _in_project() -> bool:
    if not os.path.exists("bootstrap.py"):
        return False
    if not os.path.exists(EXP_RESULTS_FOLDER):
        return False
    try:
        bootstrap = load_bootstrap()
        expected_funcs = ['foundation', 'builtin', 'user_code']
        for func in expected_funcs:
            if not hasattr(bootstrap, func):
                print(f"WARNING: Bootstrap missing expected function: {func}")
                return False
    except Exception as e:
        print(f"ERROR: Bootstrap inspection failed: {e}")
        #return True  # fallback: we assume it's okay for now
        # TODO: consider setting argument to return True vs dump.
        raise e

    return True


def is_project(folder: str) -> bool:
    """Does a lightweight inspection of a folder to determine if it
    appears to be a project. Note that a deeper inspection is performed
    by _in_project in that function inspects the bootstrap file by
    loading from within a project. However, function does not rely on
    needing to load user modules, etc so it is primarly used outside of
    a project such as getting a listing of projects.

    :param folder: The folder to inspect
    :type folder: str
    :return: True if it appears to be a project.
    :rtype: bool
    """
    if not os.path.isdir(folder):
        return False

    bootstrap_file_w_path = os.path.join(folder, "bootstrap.py")

    if not os.path.exists(bootstrap_file_w_path):
        return False

    with open(bootstrap_file_w_path, "r", encoding="utf-8") as infile:
        raw_text = infile.read()
        # it will always have a few things
        if not "foundation()" in raw_text:
            return False
        if not "builtin()" in raw_text:
            return False
        if not "user_code()" in raw_text:
            return False

    # now check for the results folder
    results_folder = os.path.join(folder, EXP_RESULTS_FOLDER)
    if not os.path.exists(results_folder):
        return False
    if not os.path.isdir(results_folder):
        return False

    return True


def _list_experiments():
    """Lists the experiments to the console.

    :raises ValueError: failed to properly parse an experiment
    """
    exp_list = project_manager.get_available_experiments()
    name_str = "experiment name"
    desc_str = "description"

    title = f"{name_str:<27} | {desc_str}"
    bar_length = len(title)
    entries_str = ""
    for idx, entry in enumerate(exp_list):
        if not isinstance(entry, AvailableExperimentModel) and \
                isinstance(entry, dict):
            try:
                entry = AvailableExperimentModel(**entry)
            except ValidationError:
                msg = ("Unexpected type from project_manager for exp "
                       f"listing. Observed: {type(entry)}")
                if _logger is not None:
                    _logger.error(msg)
                raise ValueError(msg)

        entry_str = f" {idx} : {entry.name:<22} | {entry.description}"
        if len(entry_str) > bar_length:
            bar_length = len(entry_str)
        entries_str += f"{entry_str}\n"

    bar = "="*bar_length
    output_str = f"\n{title}\n{bar}\n{entries_str}"
    print(output_str)

# --------------------------------------------------------------------
# --------------------------------------------------------------------
#          !!! MANAGER !!! THIS IS THE MAIN FUNCTION
# --------------------------------------------------------------------
# --------------------------------------------------------------------


def manager() -> None:
    """provides a CLI management"""
    global _logger

    # -------- setup ------------
    args = _parse_args()
    exp_name: Optional[str] = None
    scenario_logger: Optional[Logger] = None
    
    # Ensures current directory is part of the path
    sys.path.insert(0, os.getcwd())

    # register signal handlers
    signal.signal(signal.SIGTERM, cleanup)
    signal.signal(signal.SIGINT, cleanup)
    
    # now see if in-project
    inside_a_project = _in_project()
    try:
        # logging? only if its in a project
        if inside_a_project:
            # Save pid
            pid = os.getpid()
            with open(file=PID_FILE, mode="w", encoding="utf-8") as pidfile:
                pidfile.write(str(pid))
                pidfile.close()
            # Set up module-specific logger
            _logger = create_logger(
                "adgtk.cli.log",
                logger_name=__name__,
                subdir="framework"
            )
            # load bootstrap file
            _run_bootstrap()
        else:
            print("Not in bootstrap?")

        if args.command is None:
            print(intro)
            sys.exit(0)

        # ----- Project command ------
        if args.command == 'project':
            if args.action == "list":
                _list_projects()
                sys.exit(0)
            if args.action == "create":
                if inside_a_project:
                    msg = (
                        "WARNING: It appears you are in a project. "
                        "Cancelling request to create a project.")
                    print(msg)
                    if _logger is not None:
                        _logger.error(msg)
                    sys.exit(1)
                if args.folder is None:
                    print("Name of the project is required")
                    sys.exit()
                print(args)
                _create_project(name=args.folder)

            sys.exit(0)

        # ----- batch command ------
        elif args.command == 'batch':
            if args.name is None:
                print("Missing batch name. Name is required for batch jobs")
            experiment_runner.run_batch(filename=args.name)
            sys.exit(0)

        # ----- experiment command ------
        elif args.command == 'experiment':
            if args.action == "build":
                exp_name = None
                if args.name is not None:
                    exp_name = args.name

                experiment_builder.build_experiment(name=args.name)
                sys.exit()
            elif args.action == "list":
                _list_experiments()
                sys.exit(0)
            elif args.action == "run":
                # get the name of the experiment
                if args.name is not None:
                    exp_name = args.name

                experiment_runner.run_scenario(
                    filename=exp_name,
                    append_timestamp=False,
                    use_count=True)                
                sys.exit()

            else:
                print("unknown action")
                sys.exit()

        elif args.command == "factory":
            factory.report(group=args.group, tags=args.tags)
            sys.exit()
    except KeyboardInterrupt:
        try:        
            scenario_logger = get_scenario_logger()
            scenario_logger.info("Keyboard interupt. ending early")
        except RuntimeError:
            # if logger does not exist, it raises RuntimeError
            pass
        if os.path.exists(PID_FILE):
            os.remove(PID_FILE)

if __name__ == '__main__':
    manager()



"""Basic wrapper objects. Helps ease interopability across code base."""

from typing import Any
from enum import Enum, auto
from dataclasses import dataclass


class StateType(Enum):
    """MVP. need to evaluate which onese to add/remove"""
    TENSOR = auto()
    ARRAY = auto()
    STRING = auto()
    PRESENTABLE_RECORD = auto()
    PRESENTABLE_GROUP = auto()
    DICT = auto()
    OTHER = auto()


class ActionType(Enum):
    """MVP. need to evaluate which onese to add/remove"""
    INT = auto()
    STRING = auto()
    TENSOR = auto()
    ARRAY = auto()
    OTHER = auto()


@dataclass
class State:
    """A state object"""
    state_type: StateType         # The type of the state
    value: Any              # The value of the state
    label: Any = None       # A label for the state if one exists


@dataclass
class Action:
    """An action object"""
    value: Any              # The value of the action
    action_type: ActionType        # The type of the action



"""Built-in measurements

"""
from .factory import register_to_measurement_factory

__all__ = [
    "dict_total_str_length",
    "key_overlap",
    "string_length",    
]
# ----------------------------------------------------------------------
# String measurements
# ----------------------------------------------------------------------

@register_to_measurement_factory(tags=["string", "built-in"])
def string_length(a:str) -> int:
    """Wraps built-in python len.

    Args:
        a (str): The text to measure

    Returns:
        int: The number of characters
    """
    return len(a)


# ----------------------------------------------------------------------
# Dictionary measurements
# ----------------------------------------------------------------------

@register_to_measurement_factory(tags=["dict", "built-in"])
def dict_total_str_length(a:dict) -> int:
    """Iterates through a single dictionary and for every string it
    finds it adds that length to the overall text. if its a number it
    converts to a string in order to measure it as a string. If a value
    is a list or a dict, it also processes
    Args:
        a (dict): The dictionary to measure

    Returns:
        int: the number of characters
    """
    total = 0
    for _, item in a.items():
        try:
            if isinstance(item, str) or isinstance(item, (int, float)):
                total += len(str(item))
            elif isinstance(item, dict):
                total += dict_total_str_length(item)
            elif isinstance(item, list):
                for entry in item:
                    if isinstance(entry, str) or \
                        isinstance(entry, (int, float)):
                        total += len(str(entry))
        except (TypeError, ValueError):
            pass
    return total


# ----------------------------------------------------------------------
# String comparisons
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Dict comparisons
# ----------------------------------------------------------------------
def key_overlap(a:dict, b:dict) -> float:
    """Calculates the key overlap percentage

    Args:
        a (dict): The first entry
        b (dict): The second entry

    Returns:
        float: The ratio of overlap (with longest key set as denominator)
    """
    if not isinstance(a, dict) or not isinstance(b, dict):
        raise TypeError("Both inputs must be dictionaries")
        
    a_keys = set(a.keys())
    b_keys = set(b.keys())
    longest = max(len(a_keys), len(b_keys))    
    if longest == 0:
        return 0
    overlap = a_keys & b_keys
    return len(overlap) / longest


"""data.py is focused on the internal data structures and functions that
support the society.
"""

from typing import get_args, Literal, Optional, Protocol, Union
from pydantic import BaseModel, Field
import datetime


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
TRASH_DIR = ".trash"
INTERNAL_DATA_DIR = ".registered_files"


# ----------------------------------------------------------------------
# Foundation
# ----------------------------------------------------------------------
OrientationTypes = Literal[
     "dict_contains_list",
     "dict",
     "list_contains_dict",
     "list_contains_string",
     "pandas",
     "other"]

# what structure will be found on load? helps inform how to load.
FileEncodingTypes = Literal[
    "csv",
    "json",
    "pickle",
    "pandas",
    "text",
]

FileActionTypes=Literal["copy", "move", "none"]
PurposeTypes = Literal[
    "generated",
    "measurement",
    "messages",
    "other",    
    "performance",    
    "prompts",
]

SUPPORTED_ORIENTATION_TYPES = list(get_args(OrientationTypes))
SUPPORTED_FILE_ENCODING_TYPES = list(get_args(FileEncodingTypes))


# ----------------------------------------------------------------------
# Structure
# ----------------------------------------------------------------------

class ColumnDefinition(BaseModel):
    name: str
    type: Literal["float", "str", "dict", "int", "bool", "other"]
    description: str


class FileMetaData(BaseModel):
    """File specific. naming convention: file.csv has file.meta.json.
    This data structure is for the .meta.json."""
    description:str    
    fields: list[ColumnDefinition]
    created_by: Optional[str] = None
    updated_date: str = Field(
        default_factory=lambda: datetime.datetime.now().isoformat())


class FileDefinition(BaseModel):
    """Defines a file and how to process it. More an internal structure."""
    file_id: str        # used for indexing multiple files. must be unique
    filename:str
    path: str
    encoding: FileEncodingTypes
    # metadata: Optional[FileMetaData] = None
    metadata_file: Optional[str] = None
    tags: Optional[Union[str, list[str]]] = None


class DataDefinition(BaseModel):
    """The combined structure"""    
    shuffle_on_load: Optional[bool] = None
    key_rename_map: Optional[dict[str,str]] = None
    target_orientation: Optional[OrientationTypes] = None

class FileDataDefinition(DataDefinition):    
    file_definition: FileDefinition    


class InMemoryDataDefinition(DataDefinition):
    """Used for passing data in-memory"""
    data: Optional[Union[dict, list]]

class FileManagerConfig(BaseModel):
    """Used to make things a bit easier to maintain for the file mgr cli
    """
    label:str
    inventory_filename:str
    folder:str = INTERNAL_DATA_DIR
    load_on_init:bool=True

class FileEntry(BaseModel):
    """Used for ScenarioResults"""
    filename: str
    purpose: PurposeTypes


# ----------------------------------------------------------------------
# Protocol
# ----------------------------------------------------------------------

class CanTrackFiles(Protocol):

    def list_files(
        self,
        tag: Optional[Union[str, list[str]]] = None
    ) -> list[FileDefinition]:
        ...
    
    def get_file_id(self, filename:str) -> str: ...

    def report(self, tag:Optional[Union[str, list]]=None) -> None: ...

    def register_file(
        self,
        source_file:str,
        encoding: FileEncodingTypes,
        metadata_file: Optional[str] = None,
        tags: Optional[Union[str, list[str]]] = None,
        id:Optional[str] =None
    ) -> str:
        ...

    def retire_file(self, id:str) -> None: ...

    def get_file_definition(self, id:str) -> FileDefinition: ...



"""The Factory is a Dynamic Factory that is easy to extend object types.
The factory is a shared common resource. By having a flexible factory it
is easier to extend to new objects on behalf of the user without needing
the user to create their own factory and register that factory.

Versions:
v 0.1
- mvp

References:
- https://docs.python.org/3/library/typing.html#typing.TypedDict

TODO:

1.0

Defects:

1.0
"""

from typing import Protocol, Union, List, runtime_checkable, Callable, TypeVar
from adgtk.common import FactoryBlueprint

# ----------------------------------------------------------------------
# Exceptions
# ----------------------------------------------------------------------


class DuplicateFactoryRegistration(Exception):
    """Used for registration collision"""

    default_msg = "Registration exists. Unregister first."

    def __init__(self, message: str = default_msg):
        super().__init__(message)


class InvalidBlueprint(Exception):
    """Used when a blueprint is invalid. Covers both a FactoryBlueprint
    as well as any future blueprints."""

    default_msg = "Invalid Blueprint."

    def __init__(self, message: str = default_msg):
        super().__init__(message)


# ----------------------------------------------------------------------
# Protocols, Types, & TypeDict
# ----------------------------------------------------------------------

T = TypeVar("T", bound=Callable, covariant=True)

@runtime_checkable
class FactoryImplementable(Protocol):
    """Can be registered and created with the factory.
    """
    blueprint: FactoryBlueprint
    description: str

    __init__: Callable
    

class SupportsFactoryRegistry(Protocol):
    """A factory must have the following"""

    def registry_listing(
        self,
        group_label: Union[str, None] = None
    ) -> List[str]:
        """Lists factory entries. if no group listed it returns groups,
        else if a group is entered it returns all types in that group.

        :param group_label: A group label, defaults to None
        :type group_label: Union[str, None], optional
        :return: A listing of types in a group or all groups
        :rtype: List[str]
        """


# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------



"""CLI utilities are intende to improve overall UX"""

import logging
import os
import sys
from typing import Literal, Optional, Union
from prompt_toolkit import prompt
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text import HTML


# ----------------------------------------------------------------------
# not intended for exposing via init, etc. module specific func
# ----------------------------------------------------------------------



# ---------------------- Validation ---------------------
class MultiValidator(Validator):

    def __init__(self, validators: list):
        self.validators = validators

    def validate(self, document):
        for validator in self.validators:
            validator.validate(document)


class IntValidator(Validator):
    """Used for validating an entry is an Integer"""

    def validate(self, document):
        text = document.text

        try:
            int(text)
        except ValueError:
            raise ValidationError(
                message='Invalid entry. Input is Interger only')


class FloatValidator(Validator):
    """Used for validating an entry is an Float"""

    def validate(self, document):
        text = document.text

        try:
            float(text)
        except ValueError:
            raise ValidationError(
                message='Invalid entry. Input is Float only')


class NoWhitespaceValidator(Validator):
    """Used for validating an entry does not have a space"""

    def validate(self, document):
        text = document.text

        if " " in text:
            raise ValidationError(message='Invalid entry. Space observed')


class MinValueValidator(Validator):
    """Min value checks"""

    def __init__(self, min_value: Union[int, float]):
        self.min_value = min_value
        super().__init__()

    def validate(self, document):
        text = document.text

        try:
            val = float(text)
            if val < self.min_value:
                raise ValidationError(message="too small a value")
        except ValueError:
            # number checks is another validator
            pass


class MaxValueValidator(Validator):
    """Max value checks"""

    def __init__(self, max_value: Union[int, float]):
        self.max_value = max_value
        super().__init__()

    def validate(self, document):
        text = document.text

        try:
            val = float(text)
            if val > self.max_value:
                raise ValidationError(message="too large a value")
        except ValueError:
            # number checks is another validator
            pass


class MinLengthValidator(Validator):
    """Used for validating an entry has a minimum length"""

    def __init__(self, min_length: int):
        self.min_length = min_length
        super().__init__()

    def validate(self, document):
        text = document.text

        if len(text) < self.min_length:
            raise ValidationError(message='Invalid entry. Too short')


class MaxLengthValidator(Validator):
    """Used for validating an entry does not exceed a length"""

    def __init__(self, max_length: int):
        self.max_length = max_length
        super().__init__()

    def validate(self, document):
        text = document.text

        if len(text) > self.max_length:
            raise ValidationError(message='Invalid entry. Too long')


class ChoiceValidator(Validator):
    """Used for validating the user input is expected based on choices.
    """

    def __init__(self, choices: list):
        self.choices = choices
        choice_str = " ".join(choices)
        self.error_msg = f"Valid options are : {choice_str}"
        super().__init__()

    def validate(self, document):
        text = document.text

        if text not in self.choices:
            raise ValidationError(message=self.error_msg)


# ---------------------- UX ---------------------

def bottom_toolbar(
    helper: Union[str, None] = None,
    configuring: Union[str, None] = None,
    choices: Union[list, None] = None
) -> HTML:
    html_str = ""
    if configuring is not None:
        html_str += f"Configuring [{configuring}] "

    if helper is not None:
        html_str += f'<b><style bg="ansired">{helper}</style></b>!'

    if choices is not None:
        if len(choices) > 4:
            choice_str = ",".join(choices[0:3])
            choice_str += " ..."
        else:
            choice_str = ", ".join(choices)

        if len(choice_str) > 50:
            choice_str = choices[0]
            choice_str += " ..."

        if not html_str.endswith("."):
            html_str += "."

        html_str += ' Valid choices are : <b><style bg="ansired">'
        html_str += f'{choice_str}</style></b>'

    return HTML(html_str)


def prompt_continuation(width, line_number, wrap_count):
    if wrap_count > 0:
        return " " * (width - 3) + "-> "
    else:
        return (": ").rjust(width)

# ----------------------------------------------------------------------
# intended to be used by other, "public" funcs
# ----------------------------------------------------------------------


def get_user_input(
    user_prompt: str,
    requested: Literal["float", "str", "int", "bool", "ml-str"],
    configuring: Union[str, None] = None,
    helper: Union[str, None] = None,
    choices: Union[list, None] = None,
    allow_whitespace: bool = True,
    default_selection: Union[float, str, int, bool, None] = None,
    max_characters: Union[int, None] = None,
    min_characters: Union[int, None] = None,
    min_value: Union[int, None] = None,
    max_value: Union[int, None] = None,
    limit_ml_line_length:Optional[int] = None
) -> Union[str, int, float, bool]:

    # if default_selection is not None:
    #    request += f" [{default_selection}] : "
    # else:

    validators: list[Validator] = []
    value: Union[str, int, float, bool] = ""

    if choices is not None:
        validators.append(ChoiceValidator(choices))
    if allow_whitespace is False:
        validators.append(NoWhitespaceValidator())
    if min_value is not None:
        validators.append(MinValueValidator(min_value))
    if max_value is not None:
        validators.append(MaxValueValidator(max_value))
    if min_characters is not None:
        validators.append(MinLengthValidator(min_characters))
    if max_characters is not None:
        validators.append(MaxLengthValidator(max_characters))
    if requested == "int":
        validators.append(IntValidator())
    elif requested == "float":
        validators.append(FloatValidator())
    elif requested == "bool":
        # the list should match below
        validators.append(ChoiceValidator(["True", "False"]))

    # multi-line is a bit different of an experience
    if requested == "ml-str":
        # multi-line entries are a bit different in how its handled.
        print(user_prompt)
        line = create_line(text=user_prompt, char="-")
        if limit_ml_line_length is not None:
            if len(line) > limit_ml_line_length:
                line = create_line(text="-", modified=limit_ml_line_length)
        print(line)
        print("Press [Esc] followed by [Enter] to complete input")
        print()


        value = prompt(
            "input : ",
            prompt_continuation=prompt_continuation,
            multiline=True,
            bottom_toolbar=bottom_toolbar(
                helper=helper,
                configuring=configuring))
        # going ahead and returning since its a string and ready
        return value

    # others
    user_prompt += " : "
    if requested == "int":
        if default_selection is not None:
            value = int(
                prompt(
                    user_prompt,
                    default=str(default_selection),
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices),
                    validator=MultiValidator(validators)))
        else:
            value = int(
                prompt(
                    user_prompt,
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices),
                    validator=MultiValidator(validators)))
    elif requested == "float":
        if default_selection is not None:

            value = float(prompt(
                user_prompt,
                default=str(default_selection),
                bottom_toolbar=bottom_toolbar(
                    helper=helper,
                    configuring=configuring,
                    choices=choices),
                validator=MultiValidator(validators)))
        else:
            value = float(
                prompt(
                    user_prompt,
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices),
                    validator=MultiValidator(validators)))
    elif requested == "bool":
        choices = ["True", "False"]
        request_completer = WordCompleter(choices, ignore_case=True)

        default_str = "False"
        if default_selection is not None:
            default_str = "True" if default_selection else "False"

        value = prompt(
            user_prompt,
            default=default_str,
            completer=request_completer,
            complete_while_typing=True,
            validator=MultiValidator(validators),
            bottom_toolbar=bottom_toolbar(
                helper=helper,
                configuring=configuring,
                choices=choices)
        )

        # Normalize input and convert to boolean
        value = value.strip().lower()
        if value == "true":
            return True
        elif value == "false":
            return False
        else:
            raise ValueError("Invalid input for boolean. Expected 'True' or 'False'.")

    elif requested == "str":
        if default_selection is not None:
            if choices is not None:
                request_completer = WordCompleter(choices, ignore_case=True)
                value = prompt(
                    user_prompt,
                    default=str(default_selection),
                    completer=request_completer,
                    complete_while_typing=True,
                    validator=MultiValidator(validators),
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices))

            else:
                value = prompt(
                    user_prompt,
                    default=str(default_selection),
                    validator=MultiValidator(validators),
                    complete_while_typing=True,
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices))
        else:
            if choices is not None:
                request_completer = WordCompleter(choices, ignore_case=True)
                value = prompt(
                    user_prompt,
                    completer=request_completer,
                    complete_while_typing=True,
                    validator=MultiValidator(validators),
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices))
            else:
                value = prompt(
                    user_prompt,
                    validator=MultiValidator(validators),
                    complete_while_typing=True,
                    bottom_toolbar=bottom_toolbar(
                        helper=helper,
                        configuring=configuring,
                        choices=choices))

    else:
        if requested != "ml-str":
            logging.error(f"Unexpected type {requested}")
        return 0

    if isinstance(value, str):
        if len(value) == 0:
            if default_selection is not None:
                return default_selection
            else:
                msg = "No value entered and no default provided"
                logging.error(msg)
                print(f"ERROR: {msg}. returning ``")
                return ""

    return value


def get_more_ask(configuring: Union[str, None] = None) -> bool:
    result = get_user_input(
        configuring=configuring,
        user_prompt=f"Action [{configuring}] ",
        requested="str",
        choices=["done", "more"])

    if result == "done":
        return False

    return True


# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------


def clear_screen():
    """Clears the screen in the terminal.
    """
    if sys.platform.startswith("win"):
        os.system("cls")
    else:
        print("\033c")


def create_line(
    text: str = "",
    char: str = "=",
    modified: int = 0,
    title:Union[str, None] = None
) -> str:
    """Creates a line

    :param text: The text to underline/line. defaults to ""
    :type text: str
    :param char: The character to create line with, defaults to "="
    :type char: str, optional
    :param modified: The additional/less characters in line,
        defaults to 0
    :type modified: int
    :param title: The title to add to the line, defaults to None
    :type title: Union[str, None]
    :return: _A line using both the text length and modified value
    :rtype: str
    """
    target_length = len(text) + modified    

    if title is not None:
        tmp = [char] * ((int(target_length/2) - int(len(title)/2) - 1))
        line = "".join(tmp)
        line += f" {title.upper()} "
        line = "".join(tmp)        
        if len(tmp) < target_length:
            line += char
        return line
    else:
        tmp = [char] * target_length
        return "".join(tmp)



# mypy: ignore-errors
"""Environment base
"""
# pylint: disable=unused-argument
from typing import Protocol, runtime_checkable
from adgtk.components.base import StateType, Action, ActionType, State


@runtime_checkable
class Environment(Protocol):
    """The Environment Procotol"""

    supports_state: list[StateType]
    supports_action: list[ActionType]

    def action_space(self) -> list[Action]:
        """Gets a list of acceptable actions

        Returns:
            list[Action]: The acceptable actions
        """

    def reset(self, value: int = 0) -> tuple[State, bool]:
        """Resets the state

        :param value: The state index., defaults to 0
        :type value: int, optional
        :return: The new state, and rolling over?
        :rtype: tuple[State, bool]
        """

    def step(self, action: Action) -> tuple[State, float, bool]:
        """Take action on the environment

        :param action: The action to take
        :type action: Action
        :return: State, reward, and rollover
        :rtype: tuple[State, float, bool]
        """
